/* icode.c
	P. Chongstitvatana
	since R1						27 Aug 97
	start project RZ				3 Apr 2001
	project A1						22 Oct 2002
	project Som						13 Sept 2003
	code generator som-c			2 Jan 2004
	tail call elimination			Feb 2005
*/

#include "compile.h"

#define		PRIVATE		static

int CP = 3;				/* current code pointer */

void decode(int ip, int *code, int *data){
	*code = CS[ip] & 255;
	*data = CS[ip] >> 8;
}

void decode2(int op, int *code, int *data){
	*code = op & 255;
	*data = op >> 8;
}

void putcode(int ip, int code, int data){
	CS[ip] = (data << 8) | (code & 255);
}

PRIVATE void pl(FILE *fo, char *s1){
	fprintf(fo,"%s",s1);
}
PRIVATE void pl2(FILE *fo, char *s1, int a){
	fprintf(fo,"%s %d",s1,a);
}
PRIVATE void plnm(FILE *fo, char *s1, char *s2){
	fprintf(fo,"%s %s",s1,s2);
}
PRIVATE void prCode(FILE *fo, int ip){
	int op, a;
	char *nm, buf[80];
	decode(ip,&op,&a);
	switch(op){
	case icAdd: pl(fo,"Add"); break;
	case icSub: pl(fo,"Sub"); break;
	case icMul: pl(fo,"Mul"); break;
	case icDiv: pl(fo,"Div"); break;
	case icBand: pl(fo,"Band"); break;
	case icBor: pl(fo,"Bor"); break;
	case icBxor: pl(fo,"Bxor"); break;
	case icNot: pl(fo,"Not"); break;
	case icEq: pl(fo,"Eq"); break;
	case icNe: pl(fo,"Ne"); break;
	case icLt: pl(fo,"Lt"); break;
	case icLe: pl(fo,"Le"); break;
	case icGe: pl(fo,"Ge"); break;
	case icGt: pl(fo,"Gt"); break;
	case icShl: pl(fo,"Shl"); break;
	case icShr: pl(fo,"Shr"); break;
	case icMod: pl(fo,"Mod"); break;
	case icLdx: pl(fo,"Ldx"); break;
	case icStx: pl(fo,"Stx"); break;
	case icRet: pl2(fo,"Ret",a); break;
	case icArray: pl(fo,"Array"); break;
	case icEnd: pl(fo,"End"); break;
	case icGet: pl2(fo,"Get",a); break;
	case icPut: pl2(fo,"Put",a); break;
	case icLd: plnm(fo,"Ld",findName(tyGVAR,a)); break;
	case icSt: plnm(fo,"St",findName(tyGVAR,a)); break;
	case icJmp: pl2(fo,"Jmp",ip+a); break;
	case icJt: pl2(fo,"Jt",ip+a); break;
	case icJf: pl2(fo,"Jf",ip+a); break;
	case icLit: pl2(fo,"Lit",a); break;
	case icCall: plnm(fo,"Call",findName(tyFUNC,a)); break;
	case icInc: pl2(fo,"Inc",a); break;
	case icDec: pl2(fo,"Dec",a); break;
	case icSys: pl2(fo,"Sys",a); break;
	case icCase: pl(fo,"Case"); break;
	case icFun: plnm(fo,"Fun",findName(tyFUNC,ip)); break;
	default: pl(fo,"Undef");
	}
}
// output object code to file
void outCode(FILE *fo){
	int i;
	fprintf(fo,"%d\n",CP-1);
	for(i=1; i<CP; i++){
		fprintf(fo,"%d ",CS[i]);
		if( (i % 10 ) == 0 ) fprintf(fo,"\n");
	}
}
/*
void readCode(FILE *fi){
	int i, n, c;
	fscanf(fi,"%d",&n);
	for(i=1; i<=n; i++){
		fscanf(fi,"%d",&c);
		CS[i] = c;
	}
}
*/
// output listing to file
void showCode(FILE *fo){
	int i;
	for(i=1; i<CP; i++){
		fprintf(fo,"\n  %5d ",i);
		prCode(fo,i);
	}
	fprintf(fo,"\n");
}

// output listing for macro
/*
void showCode2(int ip, int len){
	int i;
	for(i=ip; i<ip+len; i++){
		printf("\n  %5d ",i);
		prCode(stdout,i);
		if( (CS[i] & 255) == icEnd) break;
	}
	printf("\n");
}
*/
/* out code with argument */
void outa( int code, int arg){
	CS[CP] = (arg << 8 ) | (code & 255);
	CP++;
}

/* out singleton, arg = 0 */
void outs( int code){
	CS[CP] = code;
	CP++;
}

/* change arg, preserve code */
void patch( int ads, int v ){
	CS[ads] = (v << 8) | (CS[ads] & 255);
}

// the end of body is at CP
PRIVATE void reName(int fn){
	int i, c, d, local, ref;
	local = getLv(fn);
	ref = getRef(fn);
	for( i=ref+1; i<CP; i++ ) {
		decode(i,&c, &d);
		switch( c ) {
		case icGet :
		case icPut :
		case icInc :
		case icDec : patch(i, local-d+1); break;
		}
	}
}

PRIVATE int isJmp(int c){
	return c == icJmp || c == icJt || c == icJf;
}

// change tail-call to pass and jmp
// ref is at "ret", fn is idx to current function
PRIVATE void tailcall(int ref, int fn){
	int i, ref2, c, d, c1, d1, arg;
	ref2 = getRef(fn);
	decode(ref,&c,&d);			// ret d
	decode(ref-1,&c1,&d1);		// call
	if (c1 == icCall && d1 == ref2){ // tail-call
		arg = getArg(fn);
		CP = ref - 1;
		for(i=1; i<=arg; i++) 	// pass param
			outa(icPut,arg-i+1);
		outa(icJmp,ref2-CP+1);
		outa(icRet,d);			// necessary !
		// correct jump to ref
		for(i=ref2; i<ref; i++){
			decode(i,&c,&d);
			if (isJmp(c) && (d+i) == ref)
				patch(i,CP-1-i);
		}
	}
}

// do a small improvement: short cut jump to ret, jmp to jmp
/*
PRIVATE void improv2(int ref){
	int i,c1,d1,c2,d2;
	for( i=ref+1; i<CP-1; i++ ){
		decode(i,&c1,&d1);
		if( c1==icJmp || c1==icJt || c1==icJf ){
		    decode(i+d1,&c2,&d2);
			if( c1 == icJmp && c2 == icRet )
				putcode(i,icRet,d2);
			else if( c2==icJmp )
				patch(i,d1+d2);
		}
	}
}
*/
// only when function body is not empty
void clean(int fn) {
	int ref, a;
	ref = getRef(fn);
	patchbreak();
	a = getLv(fn) + 1;
	outa(icRet,a);
	tailcall(CP-1,fn);
	// update fun header to any change of lv
	a = getLv(fn) - getArg(fn) + 1;
	putcode(ref,icFun,a);
    reName(fn);
//  improv2(ref);
//	improv2(ref);
}

// for macro, do only when body is not empty
void cleanm(int fn) {
	int i, ref;
	ref = getRef(fn);
	patchbreak();
	outs(icEnd);
//	showCode2(ref,10);
	tailcall(CP-1,fn);
	// check if any assig local, full macro
	for(i=ref; i<CP; i++){
		if( (CS[i] & 255) == icPut ){
			setType(fn,tyFULL);
			break;
		}
	}
//  improv2(ref);
//	improv2(ref);
}

// change Calli to Call from CS[a]..CS[b]
void patchCalli(int a, int b){
	int i, c, d;
	for(i=a; i<=b; i++){
		decode(i,&c,&d);
		if( c == icCalli )
			putcode(i,icCall,getRef(d));
	}
}


